# -*- coding: utf-8 -*-
"""
    eve-swagger.definitions
    ~~~~~~~~~~~~~~~~~~~~~~~
    swagger.io extension for Eve-powered REST APIs.

    :copyright: (c) 2016 by Nicola Iarocci.
    :license: BSD, see LICENSE for more details.
"""
from flask import current_app as app

from eve_swagger import OrderedDict


def definitions():
    definitions = OrderedDict()
    dr_sources = {}
    for rd in app.config['DOMAIN'].values():
        dr_sources.update(_get_dr_sources(rd['schema']))

    for rd in app.config['DOMAIN'].values():
        if rd.get('disable_documentation'):
            continue
        title = rd['item_title']
        definitions[title] = _object(rd, dr_sources)
        if 'description' in rd:
            definitions[title]['description'] = rd['description']

    # add data_relation source fields to #/definitions/
    definitions.update(dr_sources)
    return definitions


def _object(rd, dr_sources):
    props = {}
    required = []
    for field, rules in rd['schema'].items():
        if rules.get('required') is True:
            required.append(field)

        def_name = rd['item_title']+'_'+field
        props[field] = _field_props(rules, dr_sources, def_name)

        if def_name in dr_sources:
            # the current field is a source of a data_relation

            # replace None in dr_sources with the field properties
            dr_sources[def_name] = OrderedDict(props[field])

            props[field] = {'$ref': '#/definitions/{0}'.format(def_name)}

        if 'data_relation' in rules:
            # the current field is a copy of another field
            dr = rules['data_relation']
            if dr['resource'] not in app.config['DOMAIN']:
                # source of data_relation does not exist
                continue
            title = app.config['DOMAIN'][dr['resource']]['item_title']
            source_def_name = title+'_'+dr['field']
            props[field] = {
                '$ref': '#/definitions/{0}'.format(source_def_name)
            }

    field_def = {}
    field_def['type'] = 'object'
    field_def['properties'] = props
    if len(required):
        field_def['required'] = required
    return field_def


def _field_props(rules, dr_sources, prefix):
    resp = {}
    map = {
        'dict': ('object',),
        'list': ('array',),
        'objectid': ('string', 'objectid'),
        'datetime': ('string', 'date-time'),
        'float': ('number', 'float'),
        'point': ('geometry', 'Point'),
        'multipoint': ('geometry', 'MultiPoint'),
        'linestring': ('geometry', 'LineString'),
        'multilinestring': ('geometry', 'MultiLineString'),
        'polygon': ('geometry', 'Polygon'),
        'multipolygon': ('geometry', 'MultiPolygon'),
        'geometrycollection': ('geometry', 'GeometryCollection')
    }

    eve_type = rules.get('type')
    if eve_type is None:
        return resp

    if 'description' in rules:
        resp['description'] = rules['description']

    if 'example' in rules:
        resp['example'] = rules['example']

    if 'allowed' in rules:
        resp['enum'] = rules['allowed']

    if 'default' in rules:
        resp['default'] = rules['default']

    if 'minlength' in rules:
        if eve_type == 'string':
            resp['minLength'] = rules['minlength']
        elif eve_type == 'list':
            resp['minItems'] = rules['minlength']

    if 'maxlength' in rules:
        if eve_type == 'string':
            resp['maxLength'] = rules['maxlength']
        elif eve_type == 'list':
            resp['maxItems'] = rules['maxlength']

    if 'min' in rules:
        if eve_type in ['number', 'integer', 'float']:
            resp['minimum'] = rules['min']

    if 'max' in rules:
        if eve_type in ['number', 'integer', 'float']:
            resp['maximum'] = rules['max']

    if 'readonly' in rules:
        resp['readOnly'] = rules['readonly']

    if 'regex' in rules:
        resp['pattern'] = rules['regex']

    if 'nullable' in rules:
        resp['nullable'] = rules['nullable']

    type = map.get(eve_type, (eve_type,))

    resp['type'] = type[0]
    if type[0] == 'object':
        # we don't support 'valueschema' rule
        if 'schema' in rules:
            # set prefix as item_title to avoid name collisions of nested
            # fields with higher up fields
            pseudo_rd = {'item_title': prefix, 'schema': rules['schema']}
            resp.update(_object(pseudo_rd, dr_sources))
    elif type[0] == 'array':
        type = 'array'
        if 'schema' in rules:
            resp['items'] = _field_props(rules['schema'], dr_sources, prefix)
        else:
            # 'items' is mandatory for swagger, we assume it's a list of
            # strings
            resp['items'] = {'type': 'string'}
    elif type[0] == 'geometry':
        # create a pseudo object
        pseudo_rd = {
            'item_title': prefix, 'schema': {
                'type': {
                    'type': 'string',
                    'allowed': [type[1]],
                    'required': True
                },
                'coordinates': {
                    'type': 'list',
                    'required': True
                }
            }
        }
        point2D_rd = {
            'schema': {
                'type': 'integer'
            },
            'minlength': 2,
            'maxlength': 2
        }
        array_rd = {
            'schema': {
                'type': 'list'
            }
        }
        s = 'schema'
        c = 'coordinates'
        if type[1] == 'Point':
            # 1st level array of point 2D
            pseudo_rd[s][c].update(point2D_rd)
        elif type[1] in ['LineString', 'MultiPoint']:
            # 2nd level array of point 2D
            pseudo_rd[s][c].update(array_rd)
            pseudo_rd[s][c][s].update(point2D_rd)
        elif type[1] in ['MultiLineString', 'Polygon']:
            # 3rd level array of point 2D
            pseudo_rd[s][c].update(array_rd)
            pseudo_rd[s][c][s].update(array_rd)
            pseudo_rd[s][c][s][s].update(point2D_rd)
        elif type[1] in ['MultiPolygon']:
            # 4th level array of point 2D
            pseudo_rd[s][c].update(array_rd)
            pseudo_rd[s][c][s].update(array_rd)
            pseudo_rd[s][c][s][s].update(array_rd)
            pseudo_rd[s][c][s][s][s].update(point2D_rd)
        elif type[1] in ['GeometryCollection']:
            # GeometryCollection require external defs
            pass
        resp.update(_object(pseudo_rd, dr_sources))
    else:
        try:
            resp['format'] = type[1]
        except IndexError:
            pass

    return resp


def _get_dr_sources(schema):
    '''
    Returns a dict of names for sources of data_relations:

    e.g. the name field in the user resource relates to (is a copy of) the
    name field of the person resource
    => return {'person_name': None}
    '''
    dr_sources = {}
    for rules in schema.values():
        if 'data_relation' in rules:
            dr = rules['data_relation']
            if dr['resource'] not in app.config['DOMAIN']:
                # source of data_relation does not exist
                continue
            title = app.config['DOMAIN'][dr['resource']]['item_title']
            def_name = title+'_'+dr['field']
            dr_sources[def_name] = None
        elif 'schema' in rules:
            if rules.get('type') in 'dict':
                # recursively handle data_relations in subdicts
                dr_sources.update(_get_dr_sources(rules['schema']))
            elif rules.get('type') == 'list' and 'schema' in rules['schema']:
                # recursively handle data_relations in lists
                dr_sources.update(_get_dr_sources(rules['schema']['schema']))
    return dr_sources



# -*- coding: utf-8 -*-

import os

MONGO_HOST = os.environ.get('MONGO_HOST', 'localhost')
MONGO_PORT = os.environ.get('MONGO_PORT', 27017)
MONGO_USERNAME = os.environ.get('MONGO_USERNAME', 'user')
MONGO_PASSWORD = os.environ.get('MONGO_PASSWORD', 'user')
MONGO_DBNAME = os.environ.get('MONGO_DBNAME', 'evedemo')


RESOURCE_METHODS = ['GET', 'POST', 'DELETE']

ITEM_METHODS = ['GET', 'PATCH', 'PUT', 'DELETE']

people = {
    # 'title' tag used in item links.
    'item_title': 'person',

    'additional_lookup': {
        'url': 'regex("[\w]+")',
        'field': 'lastname'
    },

    'schema': {
        'firstname': {
            'type': 'string',
            'minlength': 1,
            'maxlength': 10,
        },
        'lastname': {
            'type': 'string',
            'minlength': 1,
            'maxlength': 15,
            'required': True,
            'unique': True,
        },
        'role': {
            'type': 'list',
            'allowed': ["author", "contributor", "copy"],
        },
        'location': {
            'type': 'dict',
            'schema': {
                'address': {'type': 'string'},
                'city': {'type': 'string'}
            },
        },
        'born': {
            'type': 'datetime',
        },
    }
}

works = {
    'schema': {
        'title': {
            'type': 'string',
            'required': True,
        },
        'description': {
            'type': 'string',
        },
        'owner': {
            'type': 'objectid',
            'required': True,
            'data_relation': {
                'resource': 'people',
                'embeddable': True
            },
        },
    }
}

DOMAIN = {
    'people': people,
    'works': works,
}



import sys
import os
import json

import eve
import eve_swagger
from pymongo import MongoClient
from eve_swagger.tests.test_settings import MONGO_HOST, MONGO_PORT, \
    MONGO_USERNAME, MONGO_PASSWORD, MONGO_DBNAME

if sys.version_info >= (2, 7):
    import unittest  # noqa
else:
    import unittest2 as unittest  # noqa


class TestBase(unittest.TestCase):
    def setUp(self, settings=None):
        self.this_directory = os.path.dirname(os.path.realpath(__file__))
        if settings is None:
            settings = os.path.join(self.this_directory, 'test_settings.py')

        self.setupDB()

        self.settings = settings
        self.app = eve.Eve(settings=self.settings)

        self.app.register_blueprint(eve_swagger.swagger)
        self.app.config['SWAGGER_INFO'] = {
            'title': 'Test eve-swagger',
            'version': '0.0.1'
        }

        self.test_client = self.app.test_client()
        self.domain = self.app.config['DOMAIN']

        self.swagger_doc = self.get_swagger_doc()

    def tearDown(self):
        del self.app
        self.dropDB()

    def setupDB(self):
        self.connection = MongoClient(MONGO_HOST, MONGO_PORT)
        self.connection.drop_database(MONGO_DBNAME)
        if MONGO_USERNAME:
            self.connection[MONGO_DBNAME].add_user(MONGO_USERNAME,
                                                   MONGO_PASSWORD)

    def dropDB(self):
        self.connection = MongoClient(MONGO_HOST, MONGO_PORT)
        self.connection.drop_database(MONGO_DBNAME)
        self.connection.close()

    def get_swagger_doc(self):
        r = self.test_client.get('/api-docs')
        return self.parse_response(r)

    def parse_response(self, r):
        try:
            v = json.loads(r.get_data().decode('utf-8'))
        except ValueError:
            v = None
        return v



# -*- coding: utf-8 -*-
"""
    eve-swagger.objects
    ~~~~~~~~~~~~~~~~~~~
    swagger.io extension for Eve-powered REST APIs.

    :copyright: (c) 2015 by Nicola Iarocci.
    :license: BSD, see LICENSE for more details.
"""
from eve.utils import api_prefix
from flask import request, current_app as app

import eve_swagger
from eve_swagger import OrderedDict
from .validation import validate_info


def info():
    validate_info()

    cfg = app.config[eve_swagger.INFO]

    def node(parent, cfg, key):
        value = cfg.get(key)
        if value:
            parent[key] = cfg[key]

    info = OrderedDict()
    node(info, cfg, 'title')
    node(info, cfg, 'description')
    node(info, cfg, 'termsOfService')
    node(info, cfg, 'contact')
    node(info, cfg, 'license')
    node(info, cfg, 'version')

    return info


def host():
    # TODO should probably return None if 'host' has not been set as the host
    # is optional in swagger.
    return app.config.get(eve_swagger.HOST) or request.host


def base_path():
    return api_prefix()


def schemes():
    cfg = app.config[eve_swagger.INFO]
    if 'schemes' in cfg:
        return cfg['schemes']

    scheme = request.url.split(':')[0]
    return [scheme] if scheme in ['http', 'https', 'ws', 'wss'] else None


def consumes():
    return ['application/json']


def produces():
    produces = []
    if app.config.get('XML', True):
        produces.append('application/xml')
    if app.config.get('JSON', True):
        produces.append('application/json')
    return produces if produces else None


def parameters():
    parameters = OrderedDict()
    for (resource_name, rd) in app.config['DOMAIN'].items():
        if (resource_name.endswith('_versions')
                or rd.get('disable_documentation')):
            continue

        title = rd['item_title']
        lookup_field = rd['item_lookup_field']
        if lookup_field not in rd['schema']:
            rd['schema'][lookup_field] = {'type': 'objectid'}
        eve_type = rd['schema'][lookup_field]['type']
        descr = rd['schema'][lookup_field].get('description') or ''
        example = rd['schema'][lookup_field].get('example') or ''
        if 'data_relation' in rd['schema'][lookup_field]:
            # the lookup field is a copy of another field
            dr = rd['schema'][lookup_field]['data_relation']

            # resource definition of the data relation source
            source_rd = app.config['DOMAIN'][dr['resource']]

            # schema of the data relation source field
            source_def = source_rd['schema'][dr['field']]

            # key in #/definitions/...
            source_def_name = source_rd['item_title']+'_'+dr['field']

            # copy description if necessary
            descr = descr or source_def.get('description')
            descr = descr + ' (links to {0})'.format(source_def_name)

        p = OrderedDict()
        p['in'] = 'path'
        p['name'] = title.lower() + 'Id'
        p['required'] = True
        p['description'] = descr
        p['example'] = example
        p['type'] = eve_type
        if eve_type == 'objectid':
            p['type'] = 'string'
            p['format'] = 'objectid'
        elif eve_type == 'datetime':
            p['type'] = 'string'
            p['format'] = 'date-time'
        elif eve_type == 'float':
            p['type'] = 'number'
            p['format'] = 'float'

        parameters[title+'_'+lookup_field] = p

    return parameters


def responses():
    pass


def security_definitions():
    pass


def security():
    pass


def tags():
    tags = []
    for (resource_name, rd) in app.config['DOMAIN'].items():
        if (resource_name.endswith('_versions')
                or rd.get('disable_documentation')):
            continue

        tagInfo = {"name": rd['item_title']}

        if 'description' in rd:
            tagInfo['description'] = rd['description']

        tags.append(tagInfo)
    return tags


def external_docs():
    pass



# -*- coding: utf-8 -*-
"""
    eve-swagger.swagger
    ~~~~~~~~~~~~~~~~~~~
    swagger.io extension for Eve-powered REST APIs.

    :copyright: (c) 2015 by Nicola Iarocci.
    :license: BSD, see LICENSE for more details.
"""
import re
from collections import Mapping
from flask import Blueprint, jsonify, make_response, request, \
    current_app as app
from functools import wraps

from eve_swagger import OrderedDict
from .definitions import definitions
from .objects import info, host, base_path, schemes, consumes, produces, \
    parameters, responses, security_definitions, security, tags, \
    external_docs
from .paths import paths


swagger = Blueprint('eve_swagger', __name__)
swagger.additional_documentation = OrderedDict()


def add_documentation(doc):
    _nested_update(swagger.additional_documentation, doc)


def _modify_response(f):
    @wraps(f)
    def decorated(*args, **kwargs):
        if request.method == 'OPTIONS':
            resp = app.make_default_options_response()
        else:
            resp = make_response(f(*args, **kwargs))

        # CORS
        domains = app.config.get('X_DOMAINS')
        headers = app.config.get('X_HEADERS')
        max_age = app.config.get('X_MAX_AGE')
        allow_credentials = app.config.get('X_ALLOW_CREDENTIALS')
        expose_headers = app.config.get('X_EXPOSE_HEADERS')
        origin = request.headers.get('Origin')
        if origin and domains:
            if isinstance(domains, str):
                domains = [domains]

            if headers is None:
                headers = []
            elif isinstance(headers, str):
                headers = [headers]

            if expose_headers is None:
                expose_headers = []
            elif isinstance(expose_headers, str):
                expose_headers = [expose_headers]

            allow_credentials = allow_credentials is True
            methods = app.make_default_options_response().headers.get(
                'allow', '')

            h = resp.headers
            if '*' in domains:
                h['Access-Control-Allow-Origin'] = origin
                h['Vary'] = 'Origin'
            elif any(re.match(re.escape(domain), origin)
                     for domain in domains):
                h['Access-Control-Allow-Origin'] = origin
            else:
                h['Access-Control-Allow-Origin'] = ''

            h['Access-Control-Allow-Headers'] = ', '.join(headers)
            h['Access-Control-Expose-Headers'] = ', '.join(expose_headers)
            h['Access-Control-Allow-Methods'] = methods
            h['Access-Control-Max-Age'] = str(max_age)

        return resp
    return decorated


@swagger.route('/api-docs')
@_modify_response
def index():
    def node(parent, key, value):
        if value:
            parent[key] = value

    root = OrderedDict()
    root['swagger'] = '2.0'
    node(root, 'info', info())
    node(root, 'host', host())
    node(root, 'basePath', base_path())
    node(root, 'schemes', schemes())
    node(root, 'consumes', consumes())
    node(root, 'produces', produces())
    node(root, 'paths', paths())
    node(root, 'definitions', definitions())
    node(root, 'parameters', parameters())
    node(root, 'responses', responses())
    node(root, 'securityDefinitions', security_definitions())
    node(root, 'security', security())
    node(root, 'tags', tags())
    node(root, 'externalDocs', external_docs())

    _nested_update(root, swagger.additional_documentation)

    return jsonify(root)


def _nested_update(orig_dict, new_dict):
    for key, val in new_dict.items():
        if isinstance(val, Mapping):
            tmp = _nested_update(orig_dict.get(key, {}), val)
            orig_dict[key] = tmp
        elif isinstance(val, list):
            orig_dict[key] = (orig_dict.get(key, []) + val)
        else:
            orig_dict[key] = new_dict[key]
    return orig_dict



MONGO_HOST = 'localhost'
MONGO_PORT = 27017
MONGO_USERNAME = 'test_user'
MONGO_PASSWORD = 'test_pw'
MONGO_DBNAME = 'eve_swagger_test'

TRANSPARENT_SCHEMA_RULES = True
RESOURCE_METHODS = ['GET', 'POST', 'DELETE']
ITEM_METHODS = ['GET', 'PATCH', 'PUT', 'DELETE']

DOMAIN = {
    'people': {
        'description': 'the people resource',
        'type': 'dict',
        'schema': {
            'name': {
                'type': 'string',
                'required': True,
                'unique': True,
                'description': 'the last name of the person',
                'example': 'Doe'
            },
            'job': {
                'type': 'string',
                'required': True,
                'unique': True,
                'description': 'the job of the person'
            },
            'email': {
                'type': 'string',
                'required': True,
                'regex': '^.+@.+$'
            },
            'position': {
                'type': 'string',
                'readonly': True
            },
            'relations': {
                'type': 'list',
                'schema': {
                    'type': 'dict',
                    'schema': {
                        'relation_type': {
                            'type': 'string',
                            'required': True
                        },
                        'relation': {
                            'type': 'objectid',
                            'data_relation': {
                                'resource': 'people',
                                'field': '_id'
                            }
                        }
                    }
                }
            },
            'location': {
                'type': 'point'
            }
        }
    },
    'sub_resource': {
        'description': 'A sub resource to test regex urls.',
        'url': 'people/<regex("[a-f0-9]{24}"):personid>/related',
        'personid': {
            'type': 'objectid',
            'data_relation': {
                'resource': 'people',
                'field': '_id', },
        },
        'subject': {'type': 'string'},
    },
    'disabled_resource': {
        'disable_documentation': True,
        'type': 'dict',
        'schema': {
            'field_1': {'type': 'string'}
        }
    },
    'dr_resource_1': {
        'type': 'dict',
        'item_lookup_field': 'copied_field_with_description',
        'schema': {
            'copied_field_with_description': {
                'type': 'string',
                'description': 'foobar copied_field',
                'data_relation': {
                    'resource': 'people',
                    'field': 'job'
                }
            },
        }
    },
    'dr_resource_2': {
        'type': 'dict',
        'item_lookup_field': 'copied_field_without_description',
        'schema': {
            'copied_field_without_description': {
                'type': 'string',
                'data_relation': {
                    'resource': 'people',
                    'field': 'job'
                }
            }
        }
    }
}



# -*- coding: utf-8 -*-
"""
    eve-swagger.validation
    ~~~~~~~~~~~~~~~~~~~~~~
    swagger.io extension for Eve-powered REST APIs.

    :copyright: (c) 2015 by Nicola Iarocci.
    :license: BSD, see LICENSE for more details.
"""
from eve.exceptions import ConfigException
from flask import current_app as app
from cerberus import Validator

import eve_swagger

try:
    from urllib.parse import urlparse
except ImportError:
    from urlparse import urlparse


def validate_info():
    v = Validator()
    schema = {
        'title': {'required': True, 'type': 'string'},
        'version': {'required': True, 'type': 'string'},
        'description': {'type': 'string'},
        'termsOfService': {'type': 'string'},
        'contact': {
            'type': 'dict',
            'schema': {
                'name': {'type': 'string'},
                'url': {'type': 'string', 'validator': _validate_url},
                'email': {
                    'type': 'string',
                    'regex':
                    '^[a-zA-Z0-9_.+-]+@[a-zA-Z0-9-]+\.[a-zA-Z0-9-.]+$'
                }
            }
        },
        'license': {
            'type': 'dict',
            'schema': {
                'name': {'type': 'string', 'required': True},
                'url': {'type': 'string', 'validator': _validate_url}
            }
        },
        'schemes': {
            'type': 'list',
            'schema': {
                'type': 'string'
            }
        }
    }
    if eve_swagger.INFO not in app.config:
        raise ConfigException('%s setting is required in Eve configuration.' %
                              eve_swagger.INFO)

    if not v.validate(app.config[eve_swagger.INFO], schema):
        raise ConfigException('%s is misconfigured: %s' % (
            eve_swagger.INFO, v.errors))


def _validate_url(field, value, error):
    # TODO probably too weak
    o = urlparse(value)
    if not bool(o.scheme):
        error(field, 'Invalid URL')



#!/usr/bin/env python
"""
Setup for Swagger extension for Eve powered RESTful APIs
"""
from os.path import abspath, dirname, join
from shlex import split
from setuptools import setup, find_packages
from setuptools.command.test import test as TestCommand  # noqa N812


def read_file(filename):
    """Read the contents of a file located relative to setup.py"""
    with open(join(abspath(dirname(__file__)), filename)) as thefile:
        return thefile.read()


class Tox(TestCommand):
    """Integration of tox via the setuptools ``test`` command"""
    # pylint: disable=attribute-defined-outside-init
    user_options = [('tox-args=', 'a', "Arguments to pass to tox")]

    def initialize_options(self):
        TestCommand.initialize_options(self)
        self.tox_args = None

    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        from tox import cmdline  # pylint: disable=import-error
        args = self.tox_args
        if args:
            args = split(self.tox_args)
        cmdline(args=args)


setup(
    name='Eve-Swagger',
    version='0.0.11',
    description='Swagger extension for Eve powered RESTful APIs',
    long_description=read_file('README.rst'),
    author='Nicola Iarocci',
    author_email='nicola@nicolaiarocci.com',
    url='http://github.com/pyeve/eve-swagger',
    license='BSD',
    platforms=["any"],
    packages=find_packages(),
    install_requires=read_file('requirements.txt'),
    keywords=['swagger', 'eve', 'rest', 'api'],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy'
    ],
    tests_require=['tox'],
    cmdclass={
        'test': Tox,
    },
)



# -*- coding: utf-8 -*-
"""
    eve-swagger
    ~~~~~~~~~~~
    swagger.io extension for Eve-powered REST APIs.

    :copyright: (c) 2015 by Nicola Iarocci.
    :license: BSD, see LICENSE for more details.
"""
try:
    from collections import OrderedDict
except ImportError:
    from ordereddict import OrderedDict  # noqa: F401
from .swagger import swagger, add_documentation  # noqa

INFO = 'SWAGGER_INFO'
HOST = 'SWAGGER_HOST'



import json
from jsonschema import validate, ValidationError


if __name__ == '__main__':
    import sys
    import os
    import unittest
    sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__),
                                                 '..', '..')))

from eve_swagger.tests import TestBase  # noqa: E402


class TestEveSwagger(TestBase):
    def test_api(self):
        self.app.debug = True
        r = self.test_client.get('/api-docs')
        self.assertEqual(r.status_code, 200)

    def test_json(self):
        self.app.debug = True
        r = self.test_client.get('/api-docs')
        s = r.get_data().decode('utf-8')  # raises UnicodeError
        json.loads(s)  # raises ValueError

    def test_doc_is_dict(self):
        doc = self.swagger_doc
        self.assertIsInstance(doc, dict)

    def test_info(self):
        doc = self.swagger_doc

        self.assertIn('info', doc)
        self.assertIn('title', doc['info'])
        self.assertTrue(isinstance(doc['info']['title'], u''.__class__))
        self.assertIn('version', doc['info'])
        self.assertTrue(isinstance(doc['info']['version'], u''.__class__))

    def test_paths(self):
        doc = self.swagger_doc
        url = self.domain['people']['url']
        item_title = self.domain['people']['item_title']

        self.assertIn('paths', doc)
        self.assertIsInstance(doc['paths'], dict)
        self.assertIn('/' + url, doc['paths'])
        self.assertIn('/%s/{%sId}' % (url, item_title.lower()), doc['paths'])

    def test_definitions(self):
        doc = self.swagger_doc
        item_title = self.domain['people']['item_title']

        self.assertIn('definitions', doc)
        self.assertIsInstance(doc['definitions'], dict)
        self.assertIn(item_title, doc['definitions'])
        self.assertIn('properties', doc['definitions'][item_title])
        self.assertEqual(
            set(doc['definitions'][item_title]['properties'].keys()),
            set(['name', 'job', 'email', 'position', '_id',
                 'relations', 'location']))

    def test_definitions_are_jsonschema(self):
        doc = self.swagger_doc
        item_title = self.domain['people']['item_title']

        # validate if schema is jsonschema
        try:
            validate({}, doc['definitions'][item_title])
        except ValidationError:
            # validation errors are only thrown after schema is checked
            self.assertTrue(True)

    def test_parameters_people(self):
        doc = self.swagger_doc
        item_title = self.domain['people']['item_title']
        lookup_field = self.domain['people']['item_lookup_field']

        self.assertIn('parameters', doc)
        self.assertIn(item_title + '_' + lookup_field, doc['parameters'])

    def test_resource_description(self):
        doc = self.swagger_doc
        item_title = self.domain['people']['item_title']

        self.assertEqual('the people resource',
                         doc['definitions'][item_title]['description'])

    def test_field_description(self):
        doc = self.swagger_doc
        item_title = self.domain['people']['item_title']
        props = doc['definitions'][item_title]['properties']

        self.assertEqual('the last name of the person',
                         props['name']['description'])

    def test_field_example(self):
        doc = self.swagger_doc
        item_title = self.domain['people']['item_title']
        props = doc['definitions'][item_title]['properties']

        self.assertEqual('Doe', props['name']['example'])

    def test_disabled_resource(self):
        doc = self.swagger_doc
        url = self.domain['disabled_resource']['url']
        item_title = self.domain['disabled_resource']['item_title']
        lookup_field = self.domain['disabled_resource']['item_lookup_field']

        self.assertNotIn('/' + url, doc['paths'])
        self.assertNotIn('/%s/{%sId}' % (url, item_title.lower()),
                         doc['paths'])
        self.assertNotIn(item_title, doc['definitions'])
        self.assertNotIn(item_title + '_' + lookup_field, doc['parameters'])

    def test_data_relation_source_field(self):
        doc = self.swagger_doc

        source_field = self.domain['people']['item_title'] + '_job'
        self.assertIn(source_field, doc['definitions'])

    def test_reference_to_data_relation_source_field(self):
        doc = self.swagger_doc
        people_it = self.domain['people']['item_title']
        people_props = doc['definitions'][people_it]['properties']
        dr_1_it = self.domain['dr_resource_1']['item_title']
        dr_1_props = doc['definitions'][dr_1_it]['properties']
        key = people_it + '_job'

        self.assertIn('$ref', people_props['job'])
        self.assertEqual('#/definitions/%s' % key,
                         people_props['job']['$ref'])
        self.assertIn(key, doc['definitions'])

        self.assertIn('$ref', dr_1_props['copied_field_with_description'])
        self.assertEqual('#/definitions/%s' % key,
                         dr_1_props['copied_field_with_description']['$ref'])

        key = people_it + '__id'
        people_rels_props = people_props['relations']['items']['properties']
        self.assertIn('$ref', people_rels_props['relation'])
        self.assertEqual('#/definitions/%s' % key,
                         people_rels_props['relation']['$ref'])
        self.assertIn(key, doc['definitions'])

    def test_data_relation_extended_description(self):
        doc = self.swagger_doc
        item_title = self.domain['dr_resource_1']['item_title']
        lookup_field = self.domain['dr_resource_1']['item_lookup_field']
        par = doc['parameters'][item_title + '_' + lookup_field]
        people_it = self.domain['people']['item_title']

        self.assertIn('description', par)
        self.assertEqual(
            'foobar copied_field (links to {0}_job)'.format(people_it),
            par['description'])

    def test_data_relation_copied_description(self):
        doc = self.swagger_doc
        item_title = self.domain['dr_resource_2']['item_title']
        lookup_field = self.domain['dr_resource_2']['item_lookup_field']
        par = doc['parameters'][item_title + '_' + lookup_field]
        people_it = self.domain['people']['item_title']

        self.assertIn('description', par)
        self.assertEqual(
            'the job of the person (links to {0}_job)'.format(people_it),
            par['description'])

    def test_header_parameters(self):
        doc = self.swagger_doc
        url = self.domain['people']['url']
        item_title = self.domain['people']['item_title']
        url = '/%s/{%sId}' % (url, item_title.lower())

        header_parameters = []
        # assume that header parameters are equal for PUT, PATCH, and DELETE
        for method in ['put', 'patch', 'delete']:
            for p in doc['paths'][url][method]['parameters']:
                if 'in' not in p:
                    continue
                if p['in'] == 'header':
                    if method in ['patch', 'delete']:
                        # already added in 'put'
                        self.assertIn(p, header_parameters)
                    else:
                        header_parameters += [p]

        self.assertTrue(len(header_parameters) == 1)
        h = header_parameters[0]
        self.assertIn('name', h)
        self.assertEqual(h['name'], 'If-Match')
        self.assertTrue(h['required'])

    def test_header_parameters_without_concurrency_control(self):
        url = self.domain['people']['url']
        item_title = self.domain['people']['item_title']
        url = '/%s/{%sId}' % (url, item_title.lower())

        def get_etag_param(doc):
            params = doc['paths'][url]['delete']['parameters']
            return next(p for p in params if p.get('name') == 'If-Match')

        self.app.config['IF_MATCH'] = False
        etag_param = get_etag_param(self.get_swagger_doc())
        self.assertFalse(etag_param['required'])

        self.app.config['IF_MATCH'] = True
        self.app.config['ENFORCE_IF_MATCH'] = False
        etag_param = get_etag_param(self.get_swagger_doc())
        self.assertFalse(etag_param['required'])

    def test_cors_without_origin(self):
        self.app.config['X_DOMAINS'] = ['http://example.com']
        self.app.config['X_HEADERS'] = ['Origin', 'X-Requested-With',
                                        'Content-Type', 'Accept']
        self.app.config['X_MAX_AGE'] = 2000
        r = self.test_client.get('/api-docs')

        self.assertEqual(r.status_code, 200)
        self.assertNotIn('Access-Control-Allow-Origin', r.headers)
        self.assertNotIn('Access-Control-Allow-Headers', r.headers)
        self.assertNotIn('Access-Control-Allow-Methods', r.headers)
        self.assertNotIn('Access-Control-Max-Age', r.headers)
        self.assertNotIn('Access-Control-Expose-Headers', r.headers)

    def test_cors_with_origin(self):
        self.app.config['X_DOMAINS'] = 'http://example.com'
        self.app.config['X_HEADERS'] = ['Origin', 'X-Requested-With',
                                        'Content-Type', 'Accept']
        self.app.config['X_MAX_AGE'] = 2000
        r = self.test_client.get('/api-docs',
                                 headers={'Origin': 'http://example.com'})

        self.assertEqual(r.status_code, 200)
        self.assertIn('Access-Control-Allow-Origin', r.headers)
        self.assertEqual(r.headers['Access-Control-Allow-Origin'],
                         'http://example.com')
        self.assertIn('Access-Control-Allow-Headers', r.headers)
        self.assertEqual(
            set(r.headers['Access-Control-Allow-Headers'].split(', ')),
            set(['Origin', 'X-Requested-With', 'Content-Type', 'Accept']))
        self.assertIn('Access-Control-Allow-Methods', r.headers)
        self.assertEqual(
            set(r.headers['Access-Control-Allow-Methods'].split(', ')),
            set(['HEAD', 'OPTIONS', 'GET']))
        self.assertIn('Access-Control-Max-Age', r.headers)
        self.assertEqual(r.headers['Access-Control-Max-Age'],
                         '2000')

    def test_tags(self):
        doc = self.swagger_doc
        item_title = self.domain['people']['item_title']

        self.assertIn('tags', doc)
        tag_names = [tag['name'] for tag in doc['tags']]
        self.assertTrue(len([n for n in tag_names if n == item_title]) == 1)

        url = '/' + self.domain['people']['url']
        for m in ['get', 'post', 'delete']:
            path_m = doc['paths'][url][m]
            self.assertIn('tags', path_m)
            self.assertIn(item_title, path_m['tags'])

        url = '/%s/{%sId}' % (self.domain['people']['url'], item_title.lower())
        for m in ['get', 'patch', 'put', 'delete']:
            path_m = doc['paths'][url][m]
            self.assertIn('tags', path_m)
            self.assertIn(item_title, path_m['tags'])

    def test_status_codes(self):
        doc = self.swagger_doc
        url = self.domain['people']['url']

        people = doc['paths']['/' + url]
        self.assertIn('200', people['get']['responses'])
        self.assertIn('201', people['post']['responses'])
        self.assertIn('204', people['delete']['responses'])

        item_title = self.domain['people']['item_title']
        person = doc['paths']['/%s/{%sId}' % (url, item_title.lower())]
        self.assertIn('200', person['get']['responses'])
        self.assertIn('200', person['patch']['responses'])
        self.assertIn('200', person['put']['responses'])
        self.assertIn('204', person['delete']['responses'])

    def test_schemes_override(self):
        self.app.config['SWAGGER_INFO']['schemes'] = ['https']
        r = self.test_client.get('/api-docs')
        self.assertEqual(r.status_code, 200)
        s = r.get_data().decode('utf-8')
        d = json.loads(s)
        self.assertEqual(d.get('schemes'), ['https'])

    def test_sub_resource_regex(self):
        url = self.domain['sub_resource']['url']
        resource_title = self.domain['sub_resource']['resource_title']

        self.assertEqual('people/<personid>/related', url)
        self.assertEqual('people/<personid>/related', resource_title)


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
"""
    eve-swagger.paths
    ~~~~~~~~~~~~~~~~~
    swagger.io extension for Eve-powered REST APIs.
    :copyright: (c) 2015 by Nicola Iarocci.
    :license: BSD, see LICENSE for more details.
"""
import re

from textwrap import dedent

from flask import current_app as app

from eve_swagger import OrderedDict


# TODO consider adding at least a 'schema' property to response objects
# TODO take auth into consideration

def paths():
    def _clear_regex(_str):
        return '/'.join(map(lambda x: re.sub('regex(.*):', '', x),
                            _str.split('/')))

    paths = OrderedDict()
    for resource, rd in app.config['DOMAIN'].items():
        if (rd.get('disable_documentation')
                or resource.endswith('_versions')):
            continue

        rd['url'] = _clear_regex(rd['url'])
        rd['resource_title'] = _clear_regex(rd['resource_title'])

        methods = rd['resource_methods']
        if methods:
            url = '/%s' % rd['url']
            paths[url] = _resource(resource, rd, methods)

        methods = rd['item_methods']
        if methods:
            item_id = '%sId' % rd['item_title'].lower()
            url = '/%s/{%s}' % (rd['url'], item_id)
            paths[url] = _item(resource, rd, methods)

    return paths


def _resource(resource, rd, methods):
    item = OrderedDict()
    describe_hooks = app.config.get('ENABLE_HOOK_DESCRIPTION', False)
    if 'GET' in methods:
        item['get'] = get_response(rd)
    if 'POST' in methods:
        item['post'] = post_response(rd)
    if 'DELETE' in methods:
        item['delete'] = delete_response(rd)

    if describe_hooks:
        for m in methods:
            hook_desc = _hook_descriptions(resource, m)
            if hook_desc != '':
                item[m.lower()]['description'] = '**Hooks**:' + hook_desc

    return item


def _item(resource, rd, methods):
    item = OrderedDict()
    describe_hooks = app.config.get('ENABLE_HOOK_DESCRIPTION', False)
    if 'GET' in methods:
        item['get'] = getitem_response(rd)
    if 'PUT' in methods:
        item['put'] = put_response(rd)
    if 'PATCH' in methods:
        item['patch'] = patch_response(rd)
    if 'DELETE' in methods:
        item['delete'] = deleteitem_response(rd)

    if describe_hooks:
        for m in methods:
            hook_desc = _hook_descriptions(resource, m, item=True)
            if hook_desc != '':
                item[m.lower()]['description'] = '**Hooks**:' + hook_desc

    return item


def get_ref_schema(rd):
    return {'$ref': '#/definitions/%s' % rd['item_title']}


def get_parameters(rd):
    return OrderedDict([
        ('in', 'body'),
        ('name', rd['item_title']),
        ('required', True),
        ('schema', get_ref_schema(rd)),
    ])


def get_response(rd):
    title = rd['resource_title']
    return OrderedDict([
        ('summary', 'Retrieves one or more %s' % title),
        ('responses', {'200': {
            'description': 'An array of %s' % title,
            'schema': {
                'type': 'array',
                'items': get_ref_schema(rd)}}}),
        ('tags', [rd['item_title']])
    ])


def post_response(rd):
    return OrderedDict([
        ('summary', 'Stores one or more %s' % rd['resource_title']),
        ('parameters', [get_parameters(rd)]),
        ('responses', {'201':
                       {'description': 'operation has been successful'}}),
        ('tags', [rd['item_title']])
    ])


def delete_response(rd):
    return OrderedDict([
        ('summary', 'Deletes all %s' % rd['resource_title']),
        ('responses', {'204':
                       {'description': 'operation has been successful'}}),
        ('tags', [rd['item_title']])
    ])


def getitem_response(rd):
    title = rd['item_title']
    return OrderedDict([
        ('summary', 'Retrieves a %s document' % title),
        ('responses', {
            '200': {
                'description': '%s document fetched successfully' % title,
                'schema': get_ref_schema(rd)
            },

        }),
        ('parameters', [id_parameter(rd)]),
        ('tags', [rd['item_title']])
    ])


def put_response(rd):
    title = rd['item_title']
    return OrderedDict([
        ('summary', 'Replaces a %s document' % title),
        ('responses', {
            '200': {
                'description': '%s document replaced successfully' % title
            }
        }),
        ('parameters', [id_parameter(rd),
                        get_parameters(rd),
                        header_parameters()]),
        ('tags', [rd['item_title']])
    ])


def patch_response(rd):
    title = rd['item_title']
    return OrderedDict([
        ('summary', 'Updates a %s document' % title),
        ('responses', {
            '200': {
                'description': '%s document updated successfully' % title
            }
        }),
        ('parameters', [id_parameter(rd),
                        get_parameters(rd),
                        header_parameters()]),
        ('tags', [rd['item_title']])
    ])


def deleteitem_response(rd):
    title = rd['item_title']
    return OrderedDict([
        ('summary', 'Deletes a %s document' % title),
        ('responses', {
            '204': {
                'description': '%s document deleted successfully' % title
            }
        }),
        ('parameters', [id_parameter(rd), header_parameters()]),
        ('tags', [rd['item_title']])
    ])


def id_parameter(rd):
    return {'$ref': '#/parameters/{0}_{1}'.format(rd['item_title'],
                                                  rd['item_lookup_field'])}


def header_parameters():
    r = OrderedDict()
    r['in'] = 'header'
    r['name'] = 'If-Match'
    r['description'] = 'Current value of the _etag field'
    r['required'] = app.config['IF_MATCH'] and app.config['ENFORCE_IF_MATCH']
    r['type'] = 'string'
    return r


def _hook_descriptions(resource, method, item=False):
    if method == 'GET':
        if item is True:
            events = ['on_pre_GET',
                      'on_pre_GET_' + resource,
                      'on_fetched_item',
                      'on_fetched_item_' + resource,
                      'on_post_GET',
                      'on_post_GET_' + resource]
        else:
            events = ['on_pre_GET',
                      'on_pre_GET_' + resource,
                      'on_fetched_resource',
                      'on_fetched_resource_' + resource,
                      'on_post_GET',
                      'on_post_GET_' + resource]

    if method == 'POST':
        events = ['on_pre_POST',
                  'on_pre_POST_' + resource,
                  'on_insert',
                  'on_insert_' + resource,
                  'on_inserted',
                  'on_inserted_' + resource,
                  'on_post_POST',
                  'on_post_POST_' + resource]
    if method == 'PUT':
        events = ['on_pre_PUT',
                  'on_pre_PUT_' + resource,
                  'on_replace',
                  'on_replace_' + resource,
                  'on_replaced',
                  'on_replaced_' + resource,
                  'on_post_PUT',
                  'on_post_PUT_' + resource]
    if method == 'PATCH':
        events = ['on_pre_PATCH',
                  'on_pre_PATCH_' + resource,
                  'on_update',
                  'on_update_' + resource,
                  'on_updated',
                  'on_updated_' + resource,
                  'on_post_PATCH',
                  'on_post_PATCH_' + resource]
    if method == 'DELETE':
        if item is True:
            events = ['on_pre_DELETE',
                      'on_pre_DELETE_' + resource,
                      'on_delete_item',
                      'on_delete_item_' + resource,
                      'on_deleted_item',
                      'on_deleted_item_' + resource,
                      'on_post_DELETE',
                      'on_post_DELETE_' + resource]
        else:
            events = ['on_pre_DELETE',
                      'on_pre_DELETE_' + resource,
                      'on_delete_resource',
                      'on_delete_resource_' + resource,
                      'on_deleted_resource',
                      'on_deleted_resource_' + resource,
                      'on_post_DELETE',
                      'on_post_DELETE_' + resource]

    res = ''
    for e in events:
        callbacks = getattr(app, e)
        if len(callbacks) > 0:
            res += '\n* `' + e + '`:\n\n'
        for cb in callbacks:
            if cb.__doc__:
                s = '\n    '
                s += '\n    '.join(dedent(cb.__doc__).strip().split('\n'))
                res += '  * `' + cb.__name__ + '`:\n' + s + '\n\n'
            else:
                # there is no docstring provided, still add the hook name for
                # information
                res += '  * `' + cb.__name__ + '`:\nno documentation\n\n'
    return res
